from decimal import Decimal
from enum import IntEnum
from itertools import islice

from typing_extensions import override

from aiovantage.command_client import Converter

from .base import Interface, method


class RGBLoadInterface(Interface):
    """RGB load interface."""

    interface_name = "RGBLoad"

    class RGBChannel(IntEnum):
        """The RGB color channels."""

        Red = 0
        Green = 1
        Blue = 2
        White = 3

    class HSLAttribute(IntEnum):
        """The HSL color attributes."""

        Hue = 0
        Saturation = 1
        Lightness = 2

    class ColorName(IntEnum):
        """Color."""

        Unknown = 0
        White = 1
        Red = 2
        Yellow = 3
        Green = 4
        Magenta = 5
        Blue = 6
        Cyan = 7

    # Properties
    rgb: tuple[int, int, int] | None = None
    rgbw: tuple[int, int, int, int] | None = None
    hsl: tuple[int, int, int] | None = None

    # Methods
    @method("SetRGB", "SetRGBSW", "SetRGBFollowLevel")
    async def set_rgb(
        self,
        red: int = 255,
        green: int = 255,
        blue: int = 255,
        *,
        sw: bool = False,
        follow_level: bool = False,
    ) -> None:
        """Set the color of an RGB load.

        Args:
            red: The red value of the color, (0-255)
            green: The green value of the color, (0-255)
            blue: The blue value of the color, (0-255)
            sw: Set the cached value instead of the hardware value.
            follow_level: Follow the level of the load.
        """
        # INVOKE <id> RGBLoad.SetRGB <red> <green> <blue>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGB <red> <green> <blue>
        if follow_level and sw:
            raise ValueError("Cannot set both follow_level and sw")

        if follow_level:
            method = "RGBLoad.SetRGBFollowLevel"
        elif sw:
            method = "RGBLoad.SetRGBSW"
        else:
            method = "RGBLoad.SetRGB"

        await self.invoke(method, red, green, blue)

    @method("GetRGB", "GetRGBHW")
    async def get_rgb(self, channel: RGBChannel, *, hw: bool = False) -> int:
        """Get a single RGB color channel of a load from the controller.

        Args:
            channel: The channel to get the color of.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the RGB channel, 0-255.
        """
        # INVOKE <id> RGBLoad.GetRGB <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetRGB <channel>
        return await self.invoke(
            "RGBLoad.GetRGBHW" if hw else "RGBLoad.GetRGB", channel
        )

    @method("SetHSL", "SetHSLSW")
    async def set_hsl(
        self,
        hue: int,
        saturation: float | Decimal,
        lightness: float | Decimal,
        *,
        sw: bool = False,
    ) -> None:
        """Set the color of an HSL load.

        Args:
            hue: The hue value of the color, in degrees (0-360).
            saturation: The saturation value of the color, in percent (0-100).
            lightness: The lightness value of the color, in percent (0-100).
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> RGBLoad.SetHSL <hue> <saturation> <lightness>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetHSL <hue> <saturation> <lightness>
        await self.invoke(
            "RGBLoad.SetHSLSW" if sw else "RGBLoad.SetHSL", hue, saturation, lightness
        )

    @method("GetHSL", "GetHSLHW")
    async def get_hsl(self, attribute: HSLAttribute, *, hw: bool = False) -> int:
        """Get a single HSL color attribute of a load from the controller.

        Args:
            attribute: The attribute to get the value of.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the HSL attribute, 0-360 for hue, 0-100 for saturation and
            lightness.
        """
        # INVOKE <id> RGBLoad.GetHSL <attribute>
        # -> R:INVOKE <id> <value> RGBLoad.GetHSL <attribute>
        return await self.invoke(
            "RGBLoad.GetHSLHW" if hw else "RGBLoad.GetHSL", attribute
        )

    @method("DissolveRGB", "DissolveRGBFollowLevel")
    async def dissolve_rgb(
        self,
        red: int,
        green: int,
        blue: int,
        rate: float | Decimal,
        *,
        follow_level: bool = False,
    ) -> None:
        """Transition the color of an RGB load over a number of seconds.

        Args:
            red: The new red value of the color, (0-255)
            green: The new green value of the color, (0-255)
            blue: The new blue value of the color, (0-255)
            rate: The number of seconds the transition should take.
            follow_level: Follow the level of the load.
        """
        # INVOKE <id> RGBLoad.DissolveRGB <red> <green> <blue> <rate>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveRGB <red> <green> <blue> <rate>
        if follow_level:
            method = "RGBLoad.DissolveRGBFollowLevel"
        else:
            method = "RGBLoad.DissolveRGB"

        await self.invoke(method, red, green, blue, rate)

    @method("DissolveHSL")
    async def dissolve_hsl(
        self,
        hue: int,
        saturation: float | Decimal,
        lightness: float | Decimal,
        rate: float | Decimal,
    ) -> None:
        """Transition the color of an HSL load over a number of seconds.

        Args:
            hue: The new hue value of the color, in degrees (0-360).
            saturation: The new saturation value of the color, in percent (0-100).
            lightness: The new lightness value of the color, in percent (0-100).
            rate: The number of seconds the transition should take.
        """
        # INVOKE <id> RGBLoad.DissolveHSL <hue> <saturation> <lightness> <rate>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveHSL <hue> <saturation> <lightness> <rate>
        await self.invoke("RGBLoad.DissolveHSL", hue, saturation, lightness, rate)

    @method("SetDissolveRate", "SetDissolveRateSW")
    async def set_dissolve_rate(
        self, rate: float | Decimal, *, sw: bool = False
    ) -> None:
        """Set the default dissolve rate for RGB and HSL transitions.

        Args:
            rate: The number of seconds the transition should take.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> RGBLoad.SetDissolveRate <rate>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetDissolveRate <rate>
        await self.invoke(
            "RGBLoad.SetDissolveRateSW" if sw else "RGBLoad.SetDissolveRate", rate
        )

    @method("GetDissolveRate", "GetDissolveRateHW")
    async def get_dissolve_rate(self, *, hw: bool = False) -> Decimal:
        """Get the default dissolve rate for RGB and HSL transitions.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The number of seconds the transition should take.
        """
        # INVOKE <id> RGBLoad.GetDissolveRate
        # -> R:INVOKE <id> <rate> RGBLoad.GetDissolveRate
        return await self.invoke(
            "RGBLoad.GetDissolveRateHW" if hw else "RGBLoad.GetDissolveRate"
        )

    @method("IncrementRGBComponent")
    async def increment_rgb_component(self, channel: RGBChannel) -> None:
        """Increment a single RGB color channel of a load.

        Args:
            channel: The channel to increment the color of.
        """
        # INVOKE <id> RGBLoad.IncrementRGBComponent <channel>
        # -> R:INVOKE <id> <rcode> RGBLoad.IncrementRGBComponent <channel>
        await self.invoke("RGBLoad.IncrementRGBComponent", channel)

    @method("DecrementRGBComponent")
    async def decrement_rgb_component(self, channel: RGBChannel) -> None:
        """Decrement a single RGB color channel of a load.

        Args:
            channel: The channel to decrement the color of.
        """
        # INVOKE <id> RGBLoad.DecrementRGBComponent <channel>
        # -> R:INVOKE <id> <rcode> RGBLoad.DecrementRGBComponent <channel>
        await self.invoke("RGBLoad.DecrementRGBComponent", channel)

    @method("SetRGBComponent")
    async def set_rgb_component(self, channel: RGBChannel, value: int) -> None:
        """Set a single RGB(W) color channel of a load.

        Args:
            channel: The channel to set the color of.
            value: The value to set the channel to, 0-255.
        """
        # INVOKE <id> RGBLoad.SetRGBComponent <channel> <value>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGBComponent <channel> <value>
        await self.invoke("RGBLoad.SetRGBComponent", channel, value)

    @method("IncrementHSLAttribute")
    async def increment_hsl_attribute(self, attribute: HSLAttribute) -> None:
        """Increment a single HSL color attribute of a load.

        Args:
            attribute: The attribute to increment the value of.
        """
        # INVOKE <id> RGBLoad.IncrementHSLAttribute <attribute>
        # -> R:INVOKE <id> <rcode> RGBLoad.IncrementHSLAttribute <attribute>
        await self.invoke("RGBLoad.IncrementHSLAttribute", attribute)

    @method("DecrementHSLAttribute")
    async def decrement_hsl_attribute(self, attribute: HSLAttribute) -> None:
        """Decrement a single HSL color attribute of a load.

        Args:
            attribute: The attribute to decrement the value of.
        """
        # INVOKE <id> RGBLoad.DecrementHSLAttribute <attribute>
        # -> R:INVOKE <id> <rcode> RGBLoad.DecrementHSLAttribute <attribute>
        await self.invoke("RGBLoad.DecrementHSLAttribute", attribute)

    @method("SetHSLAttribute")
    async def set_hsl_attribute(self, attribute: HSLAttribute, value: int) -> None:
        """Set a single HSL color attribute of a load.

        Args:
            attribute: The attribute to set the value of.
            value: The value to set the attribute to, 0-360 for hue, 0-100 for
                saturation and lightness.
        """
        # INVOKE <id> RGBLoad.SetHSLAttribute <attribute> <value>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetHSLAttribute <attribute> <value>
        await self.invoke("RGBLoad.SetHSLAttribute", attribute, value)

    @method("Stop")
    async def stop(self) -> None:
        """Stop the transition."""
        # INVOKE <id> RGBLoad.Stop
        # -> R:INVOKE <id> <rcode> RGBLoad.Stop
        await self.invoke("RGBLoad.Stop")

    @method("NextPreset")
    async def next_preset(self) -> None:
        """Change to the next lighting preset."""
        # INVOKE <id> RGBLoad.NextPreset
        # -> R:INVOKE <id> <rcode> RGBLoad.NextPreset
        await self.invoke("RGBLoad.NextPreset")

    @method("PreviousPreset")
    async def previous_preset(self) -> None:
        """Change to the previous lighting preset."""
        # INVOKE <id> RGBLoad.PreviousPreset
        # -> R:INVOKE <id> <rcode> RGBLoad.PreviousPreset
        await self.invoke("RGBLoad.PreviousPreset")

    @method("NextEffect")
    async def next_effect(self) -> None:
        """Change to the next lighting effect."""
        # INVOKE <id> RGBLoad.NextEffect
        # -> R:INVOKE <id> <rcode> RGBLoad.NextEffect
        await self.invoke("RGBLoad.NextEffect")

    @method("PreviousEffect")
    async def previous_effect(self) -> None:
        """Change to the previous lighting effect."""
        # INVOKE <id> RGBLoad.PreviousEffect
        # -> R:INVOKE <id> <rcode> RGBLoad.PreviousEffect
        await self.invoke("RGBLoad.PreviousEffect")

    @method("SetPreset", "SetPresetSW")
    async def set_preset(self, index: int, *, sw: bool = False) -> None:
        """Change to a specific lighting preset.

        Args:
            index: The index of the preset to change to.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> RGBLoad.SetPreset <index>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetPreset <index>
        await self.invoke("RGBLoad.SetPresetSW" if sw else "RGBLoad.SetPreset", index)

    @method("GetPreset", "GetPresetHW")
    async def get_preset(self, *, hw: bool = False) -> int:
        """Get the current lighting preset.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The index of the current preset.
        """
        # INVOKE <id> RGBLoad.GetPreset
        # -> R:INVOKE <id> <index> RGBLoad.GetPreset
        return await self.invoke("RGBLoad.GetPresetHW" if hw else "RGBLoad.GetPreset")

    @method("SetEffect", "SetEffectSW")
    async def set_effect(self, index: int, *, sw: bool = False) -> None:
        """Change to a specific lighting effect.

        Args:
            index: The index of the effect to change to.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> RGBLoad.SetEffect <index>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetEffect <index>
        await self.invoke("RGBLoad.SetEffectSW" if sw else "RGBLoad.SetEffect", index)

    @method("GetEffect", "GetEffectHW")
    async def get_effect(self, *, hw: bool = False) -> int:
        """Get the current lighting effect.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The index of the current effect.
        """
        # INVOKE <id> RGBLoad.GetEffect
        # -> R:INVOKE <id> <index> RGBLoad.GetEffect
        return await self.invoke("RGBLoad.GetEffectHW" if hw else "RGBLoad.GetEffect")

    @method("SetColorByName")
    async def set_color_by_name(self, color: ColorName) -> None:
        """Set the color of an RGB load by name.

        Args:
            color: The name of the color to set the load to.
        """
        # INVOKE <id> RGBLoad.SetColorByName <color>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetColorByName <color>
        await self.invoke("RGBLoad.SetColorByName", color)

    @method("GetColorName")
    async def get_color_name(self) -> ColorName:
        """Get the name of the color of a load from the controller.

        Returns:
            The name of the color.
        """
        # INVOKE <id> RGBLoad.GetColorName
        # -> R:INVOKE <id> <color> RGBLoad.GetColorName
        return await self.invoke("RGBLoad.GetColorName")

    @method("GetColor", "GetColorHW")
    async def get_color(self, *, hw: bool = False) -> int:
        """Get the RGB/RGBW color of a load from the controller.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The RGB(W) value of the color as a packed big-endian integer.
        """
        # To unpack the response:
        # response.to_bytes(4, byteorder='big', signed=True)
        # NOTE: The W value always seems to be 0, even for RGBW loads.

        # INVOKE <id> RGBLoad.GetColor
        # -> R:INVOKE <id> <color> RGBLoad.GetColor
        return await self.invoke("RGBLoad.GetColorHW" if hw else "RGBLoad.GetColor")

    @method("SetRGBW", "SetRGBWSW", "SetRGBWFollowLevel")
    async def set_rgbw(
        self,
        red: int = 255,
        green: int = 255,
        blue: int = 255,
        white: int = 255,
        *,
        sw: bool = False,
        follow_level: bool = False,
    ) -> None:
        """Set the color of an RGBW load.

        Args:
            red: The red value of the color, (0-255)
            green: The green value of the color, (0-255)
            blue: The blue value of the color, (0-255)
            white: The white value of the color, (0-255)
            sw: Set the cached value instead of the hardware value.
            follow_level: Follow the level of the load.
        """
        # INVOKE <id> RGBLoad.SetRGBW <red> <green> <blue> <white>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGBW <red> <green> <blue> <white>
        if follow_level and sw:
            raise ValueError("Cannot set both follow_level and sw")

        if follow_level:
            method = "RGBLoad.SetRGBWFollowLevel"
        elif sw:
            method = "RGBLoad.SetRGBWSW"
        else:
            method = "RGBLoad.SetRGBW"

        await self.invoke(method, red, green, blue, white)

    @method("GetRGBW", "GetRGBWHW")
    async def get_rgbw(self, channel: int, *, hw: bool = False) -> int:
        """Get a single RGBW color channel of a load from the controller.

        Args:
            channel: The channel to get the color of.
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the RGBW channel, 0-255.
        """
        # INVOKE <id> RGBLoad.GetRGBW <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetRGBW <channel>
        return await self.invoke(
            "RGBLoad.GetRGBWHW" if hw else "RGBLoad.GetRGBW", channel
        )

    @method("GetTransitionLevel")
    async def get_transition_level(self) -> Decimal:
        """Get the transition level of a load.

        Returns:
            The transition level of the load.
        """
        # INVOKE <id> RGBLoad.GetTransitionLevel
        # -> R:INVOKE <id> <level> RGBLoad.GetTransitionLevel
        return await self.invoke("RGBLoad.GetTransitionLevel")

    @method("DissolveRGBW", "DissolveRGBWFollowLevel")
    async def dissolve_rgbw(
        self,
        red: int,
        green: int,
        blue: int,
        white: int,
        rate: float | Decimal,
        *,
        follow_level: bool = False,
    ) -> None:
        """Transition the color of an RGBW load over a number of seconds.

        Args:
            red: The new red value of the color, (0-255)
            green: The new green value of the color, (0-255)
            blue: The new blue value of the color, (0-255)
            white: The new white value of the color, (0-255)
            rate: The number of seconds the transition should take.
            follow_level: Follow the level of the load.
        """
        # INVOKE <id> RGBLoad.DissolveRGBW <red> <green> <blue> <white> <rate>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveRGBW <red> <green> <blue> <white> <rate>
        if follow_level:
            method = "RGBLoad.DissolveRGBWFollowLevel"
        else:
            method = "RGBLoad.DissolveRGBW"

        await self.invoke(method, red, green, blue, white, rate)

    # Convenience functions, not part of the interface
    async def get_rgb_color(self) -> tuple[int, int, int]:
        """Get the RGB color of a load from the controller.

        Returns:
            The value of the RGB color as a tuple of (red, green, blue).
        """
        return tuple[int, int, int](
            [await self.get_rgb(chan) for chan in islice(self.RGBChannel, 3)]
        )

    async def get_rgbw_color(self) -> tuple[int, int, int, int]:
        """Get the RGBW color of a load from the controller.

        Returns:
            The value of the RGBW color as a tuple of (red, green, blue, white).
        """
        return tuple[int, int, int, int](
            [await self.get_rgbw(chan) for chan in self.RGBChannel]
        )

    async def get_hsl_color(self) -> tuple[int, int, int]:
        """Get the HSL color of a load from the controller.

        Returns:
            The value of the HSL color as a tuple of (hue, saturation, lightness).
        """
        return tuple[int, int, int](
            [await self.get_hsl(attr) for attr in self.HSLAttribute]
        )

    # The rgb, hsl, and rgbw properties are "virtual" properties that are populated
    # by fetching the individual color channels/attributes.
    # For status updates, we need to wait for all channels/attributes to be received
    # before updating the property.

    @override
    async def fetch_state(self) -> list[str]:
        # Fetch state from other interfaces first
        props_changed = await super().fetch_state()

        # Define the getters for our "virtual" properties
        getters = {
            "rgb": self.get_rgb_color,
            "hsl": self.get_hsl_color,
            "rgbw": self.get_rgbw_color,
        }

        # Fetch the properties
        for prop, getter in getters.items():
            props_changed.extend(self.update_properties({prop: await getter()}))

        return props_changed

    @override
    def handle_object_status(self, method: str, result: str, *args: str) -> list[str]:
        # Define the methods and the number of channels/attributes they return
        methods = {
            "RGBLoad.GetRGB": ("rgb", 3),
            "RGBLoad.GetHSL": ("hsl", 3),
            "RGBLoad.GetRGBW": ("rgbw", 4),
        }

        # Check if the method is one we're interested in
        if method not in methods:
            return super().handle_object_status(method, result, *args)

        # Get the attribute and number of channels/attributes
        attr, num_channels = methods[method]

        # Ignore channels that are out of range
        channel = Converter.deserialize(int, args[0])
        if channel not in range(num_channels):
            return []

        # Cache the value
        self._cache = getattr(self, "_cache", [0, 0, 0, 0])
        self._cache[channel] = Converter.deserialize(int, result)

        # Update the property only if all channels have been received
        if channel == num_channels - 1:
            new_value = tuple(self._cache[:num_channels])
            return self.update_properties({attr: new_value})

        return []



"""Base class for DIN station objects."""

from dataclasses import dataclass, field

from .station_object import StationObject


@dataclass(kw_only=True)
class DINStation(StationObject):
    """Base class for DIN station objects."""

    @dataclass
    class DINEnclosure:
        enclosure: int
        position: int = field(metadata={"type": "Attribute"})
        row: int = field(metadata={"type": "Attribute"})

    din_enclosure: DINEnclosure | None = field(
        default=None,
        metadata={
            "name": "DINEnclosure",
        },
    )
    module_count: int | None = None



"""Controllers for managing collections of objects from the Vantage system.

Controllers are responsible for fetching objects, fetching their state, and
keeping the state up to date.

Some controllers manage a single type of object, while others handle multiple types.
In controllers where multiple object types are managed, they are typically conceptually
related through shared behavior rather than a strict class hierarchy. These types of
controllers typically hold objects which implement one or more common interfaces. For
examples, objects in the [`BlindsController`][aiovantage.controllers.BlindsController]
all implement the [`BlindInterface`][aiovantage.object_interfaces.BlindInterface].

Controllers implement [`QuerySet`][aiovantage.controllers.QuerySet], which
provides a number of methods for filtering and finding objects, such as
[`filter`][aiovantage.controllers.QuerySet.filter] and
[`get`][aiovantage.controllers.QuerySet.get].

Controllers also implement [`EventDispatcher`][aiovantage.events.EventDispatcher],
which allows you to subscribe to events related to the objects managed by the controller
with [`subscribe`][aiovantage.events.EventDispatcher.subscribe]. The following
events are emitted by controllers:
[`ObjectAdded`][aiovantage.events.ObjectAdded],
[`ObjectUpdated`][aiovantage.events.ObjectUpdated],
[`ObjectDeleted`][aiovantage.events.ObjectDeleted]

An instance of every controller type is made available through the main
[Vantage][aiovantage.Vantage] object.
"""

from ._controllers.anemo_sensors import AnemoSensorsController
from ._controllers.areas import AreasController
from ._controllers.back_boxes import BackBoxesController
from ._controllers.base import Controller, StatusType
from ._controllers.blind_groups import BlindGroupsController, BlindGroupTypes
from ._controllers.blinds import BlindsController, BlindTypes
from ._controllers.buttons import ButtonsController
from ._controllers.dry_contacts import DryContactsController
from ._controllers.gmem import GMemController
from ._controllers.light_sensors import LightSensorsController
from ._controllers.load_groups import LoadGroupsController
from ._controllers.loads import LoadsController
from ._controllers.masters import MastersController
from ._controllers.modules import ModulesController
from ._controllers.omni_sensors import OmniSensorsController
from ._controllers.port_devices import PortDevicesController
from ._controllers.power_profiles import PowerProfilesController
from ._controllers.query import QuerySet
from ._controllers.rgb_loads import RGBLoadsController, RGBLoadTypes
from ._controllers.stations import StationsController
from ._controllers.tasks import TasksController
from ._controllers.temperatures import TemperaturesController
from ._controllers.thermostats import ThermostatsController, ThermostatTypes

__all__ = [
    "AnemoSensorsController",
    "AreasController",
    "BackBoxesController",
    "Controller",
    "BlindGroupsController",
    "BlindGroupTypes",
    "BlindsController",
    "BlindTypes",
    "ButtonsController",
    "DryContactsController",
    "GMemController",
    "LightSensorsController",
    "LoadGroupsController",
    "LoadsController",
    "MastersController",
    "ModulesController",
    "OmniSensorsController",
    "PortDevicesController",
    "PowerProfilesController",
    "QuerySet",
    "RGBLoadsController",
    "RGBLoadTypes",
    "StationsController",
    "StatusType",
    "TasksController",
    "TemperaturesController",
    "ThermostatsController",
    "ThermostatTypes",
]



from decimal import Decimal
from enum import IntEnum

from typing_extensions import override

from .base import Interface, method


class LoadInterface(Interface):
    """Load object interface."""

    interface_name = "Load"

    class RampType(IntEnum):
        """Load ramp type."""

        Up = 5
        Down = 4
        Opposite = 3
        Stop = 2
        Fixed = 6
        Variable = 7
        Adjust = 8

    class AlertState(IntEnum):
        """Load alert state."""

        Clear = 0
        Overload = 1
        BulbChange = 2
        WrongType = 3
        DCCurrent = 4
        ShortCircuit = 5

    class DimmingConfig(IntEnum):
        """Load dimming config."""

        Manual = 0
        Forward = 1
        Reverse = 2
        Auto = 3

    # Properties
    level: Decimal | None = None

    # Methods
    @method("SetLevel", "SetLevelSW")
    async def set_level(self, level: float | Decimal, *, sw: bool = False) -> None:
        """Set the level of a load.

        Args:
            level: The level to set the load to (0-100).
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Load.SetLevel <level (0-100)>
        # -> R:INVOKE <id> <rcode> Load.SetLevel <level (0-100)>
        await self.invoke("Load.SetLevelSW" if sw else "Load.SetLevel", level)

    @method("GetLevel", "GetLevelHW", property="level")
    async def get_level(self, *, hw: bool = False) -> Decimal:
        """Get the level of a load.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetLevel
        # -> R:INVOKE <id> <level (0.000-100.000)> Load.GetLevel
        return await self.invoke("Load.GetLevelHW" if hw else "Load.GetLevel")

    @method("Ramp")
    async def ramp(
        self, cmd: RampType, ramptime: float | Decimal, finallevel: float | Decimal
    ) -> None:
        """Ramp a load to a level over a number of seconds.

        Args:
            cmd: The type of ramp to perform.
            ramptime: The number of seconds to ramp the load over.
            finallevel: The level to ramp the load to (0-100).
        """
        # INVOKE <id> Load.Ramp <cmd> <time> <level>
        # -> R:INVOKE <id> <rcode> Load.Ramp <cmd> <time> <level>
        await self.invoke("Load.Ramp", cmd, ramptime, finallevel)

    @method("SetProfile")
    async def set_profile(self, profile: int) -> None:
        """Set the id of the power profile used by this load.

        Args:
            profile: The power profile id to set the load to.
        """
        # INVOKE <id> Load.SetProfile <profile>
        # -> R:INVOKE <id> <rcode> Load.SetProfile <profile>
        await self.invoke("Load.SetProfile", profile)

    @method("GetProfile")
    async def get_profile(self) -> int:
        """Get the id of the power profile used by this load.

        Returns:
            The power profile id used by the load.
        """
        # INVOKE <id> Load.GetProfile
        # -> R:INVOKE <id> <profile> Load.GetProfile
        return await self.invoke("Load.GetProfile")

    @method("GetOverrideLevel")
    async def get_override_level(self) -> Decimal:
        """Get the override level of a load.

        Returns:
            The override level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetOverrideLevel
        # -> R:INVOKE <id> <level (0.000-100.000)> Load.GetOverrideLevel
        return await self.invoke("Load.GetOverrideLevel")

    @method("RampAutoOff")
    async def ramp_auto_off(
        self,
        cmd: RampType,
        ramptime: float | Decimal,
        finallevel: float | Decimal,
        offcmd: RampType,
        offramptime: float | Decimal,
        offtimeout: float | Decimal,
    ) -> None:
        """Ramp a load to a level over a number of seconds, then ramp off after a timeout.

        Args:
            cmd: The type of ramp to perform.
            ramptime: The number of seconds to ramp the load over.
            finallevel: The level to ramp the load to (0-100).
            offcmd: The type of ramp to perform to turn the load off.
            offramptime: The number of seconds to ramp the load off over.
            offtimeout: The number of seconds to wait before turning the load off.
        """
        # INVOKE <id> Load.RampAutoOff <cmd> <time> <level> <offcmd> <offtime> <offlevel>
        # -> R:INVOKE <id> <rcode> Load.RampAutoOff <cmd> <time> <level> <offcmd> <offtime> <offlevel>
        await self.invoke(
            "Load.RampAutoOff",
            cmd,
            ramptime,
            finallevel,
            offcmd,
            offramptime,
            offtimeout,
        )

    @method("GetAlertState")
    async def get_alert_state(self) -> AlertState:
        """Get the alert state of a load.

        Returns:
            The alert state of the load.
        """
        # INVOKE <id> Load.GetAlertState
        # -> R:INVOKE <id> <alert state> Load.GetAlertState
        return await self.invoke("Load.GetAlertState")

    @method("SetAlertStateSW")
    async def set_alert_state(self, alert_state: AlertState) -> None:
        """Set the cached alert state of a load.

        Args:
            alert_state: The alert state to set the load to.
        """
        # INVOKE <id> Load.SetAlertStateSW <alert state>
        # -> R:INVOKE <id> <rcode> Load.SetAlertStateSW <alert state>
        await self.invoke("Load.SetAlertStateSW", alert_state)

    @method("GetDimmingConfig")
    async def get_dimming_config(self) -> DimmingConfig:
        """Get the dimming configuration of a load.

        Returns:
            The dimming configuration of the load.
        """
        # INVOKE <id> Load.GetDimmingConfig
        # -> R:INVOKE <id> <dimming config> Load.GetDimmingConfig
        return await self.invoke("Load.GetDimmingConfig")

    # Convenience functions, not part of the interface
    async def turn_on(
        self, transition: float | None = None, level: float | None = None
    ) -> None:
        """Turn on a load with an optional transition time.

        Args:
            transition: The time in seconds to transition to the new level, defaults to immediate.
            level: The level to set the load to (0-100), defaults to 100.
        """
        if level is None:
            level = 100

        if transition is None:
            return await self.set_level(level)

        await self.ramp(self.RampType.Fixed, transition, level)

    async def turn_off(self, transition: float | None = None) -> None:
        """Turn off a load with an optional transition time.

        Args:
            transition: The time in seconds to ramp the load down, defaults to immediate.
        """
        if transition is None:
            return await self.set_level(0)

        await self.ramp(self.RampType.Fixed, transition, 0)

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)

    @override
    def handle_category_status(self, category: str, *args: str) -> list[str]:
        if category == "LOAD":
            # STATUS LOAD
            # -> S:LOAD <id> <level (0-100)>
            return self.update_properties({"level": Decimal(args[0])})

        return super().handle_category_status(category, *args)



"""Common types."""

from dataclasses import dataclass, field
from decimal import Decimal


@dataclass(kw_only=True)
class Parent:
    """Vantage parent type."""

    vid: int
    position: int = field(metadata={"type": "Attribute"})


@dataclass(kw_only=True)
class Array:
    """Vantage array type."""

    @dataclass(kw_only=True)
    class StringData:
        string: str = ""
        size: int = field(metadata={"name": "size", "type": "Attribute"})

    @dataclass(kw_only=True)
    class BytesData:
        bytes: str = ""
        size: int = field(metadata={"name": "size", "type": "Attribute"})

    val: int | Decimal | None = field(default=None, metadata={"name": "val"})
    string: StringData | None = field(default=None, metadata={"name": "string"})
    bytes: BytesData | None = field(default=None, metadata={"name": "bytes"})
    array: "Array | None" = field(default=None, metadata={"name": "array"})



"""Example of using the get_objects helper to dump all objects of a given type."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage.config_client import ConfigClient, ConfigurationInterface

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    async with ConfigClient(args.host, args.username, args.password) as client:
        # Dump all Areas using the get_objects helper
        print("# Vantage Areas")
        async for area in ConfigurationInterface.get_objects(client, "Area"):
            print(area)
        print()

        # Dump all Loads using the get_objects helper
        print("# Vantage Loads")
        async for load in ConfigurationInterface.get_objects(client, "Load"):
            print(load)
        print()

        # Dump some StationObjects using the get_objects helper
        print("# Vantage Stations")
        async for station in ConfigurationInterface.get_objects(
            client, "Keypad", "EqCtrl"
        ):
            print(station)
        print()


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Field definitions for various "interface provided" object fields."""

from dataclasses import dataclass, field


@dataclass(kw_only=True)
class ShadeOrientation:
    """Shade orientation field support."""

    shade_orientation: str | None = field(
        default=None,
        repr=False,
        metadata={"type": "Attribute"},
    )


@dataclass(kw_only=True)
class ShadeType:
    """Shade type field support."""

    shade_type: str | None = field(
        default=None,
        repr=False,
        metadata={"type": "Attribute"},
    )



"""Base class for parent device objects."""

from dataclasses import dataclass

from .custom_device import CustomDevice


@dataclass(kw_only=True)
class ParentDevice(CustomDevice):
    """Base class for parent device objects."""



from aiovantage.objects import (
    BlindGroup,
    QISBlind,
    QubeBlind,
    RelayBlind,
    SomfyRS485ShadeChild,
    SomfyURTSI2ShadeChild,
)

from .base import Controller
from .query import QuerySet

BlindTypes = (
    QISBlind | QubeBlind | RelayBlind | SomfyRS485ShadeChild | SomfyURTSI2ShadeChild
)
"""Types managed by the blinds controller."""


class BlindsController(Controller[BlindTypes]):
    """Blinds controller."""

    vantage_types = (
        "QISBlind",
        "QubeBlind",
        "RelayBlind",
        "Somfy.RS-485_Shade_CHILD",
        "Somfy.URTSI_2_Shade_CHILD",
    )

    def in_blind_group(self, blind_group: BlindGroup) -> QuerySet[BlindTypes]:
        """Return a queryset of all loads in the given blind group."""
        return self.filter(lambda load: load.vid in blind_group.blind_table)



"""Light sensor object."""

from dataclasses import dataclass

from aiovantage.object_interfaces import LightSensorInterface, SensorInterface

from .sensor import Sensor
from .types import Parent


@dataclass(kw_only=True)
class LightSensor(Sensor, SensorInterface, LightSensorInterface):
    """Light sensor object."""

    parent: Parent



"""Play Jingle Bells on the sounder of a Vantage keypad."""

import argparse
import asyncio

from aiovantage import Vantage
from aiovantage.object_interfaces import SounderInterface

JINGLE_BELLS = [
    ("E", 0.25),
    ("E", 0.25),
    ("E", 0.5),
    ("E", 0.25),
    ("E", 0.25),
    ("E", 0.5),
    ("E", 0.25),
    ("G", 0.25),
    ("C", 0.25),
    ("D", 0.25),
    ("E", 1),
    ("F", 0.25),
    ("F", 0.25),
    ("F", 0.5),
    ("F", 0.125),
    ("F", 0.125),
    ("E", 0.25),
    ("E", 0.5),
    ("E", 0.125),
    ("E", 0.125),
    ("E", 0.125),
    ("D", 0.25),
    ("D", 0.25),
    ("E", 0.25),
    ("D", 0.5),
    ("G", 0.5),
]

FREQUENCIES = {
    "C": 261.63,
    "D": 293.66,
    "E": 329.63,
    "F": 349.23,
    "G": 392.00,
    "A": 440.00,
    "B": 493.88,
}

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("id", help="keypad id to play sound on")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main() -> None:
    """Play Jingle Bells on the sounder of a Vantage keypad."""
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Grab the keypad object
        keypad = await vantage.stations.aget(args.id)

        # Check keypad was found
        if keypad is None:
            raise ValueError(f"Keypad {args.id} not found")

        # Attach a command client, so we can use object interfaces
        keypad.command_client = vantage.command_client

        # Check the keypad object has a sounder interface
        if not isinstance(keypad, SounderInterface):
            raise ValueError(f"Object {keypad} does not have a sounder interface")

        # Turn on the sounder
        await SounderInterface.turn_on(keypad)

        # Play each note in Jingle Bells
        for note, duration in JINGLE_BELLS:
            frequency = FREQUENCIES[note]
            await keypad.set_frequency(frequency)
            await asyncio.sleep(duration)

        # Turn off the sounder
        await keypad.turn_off()


asyncio.run(main())



from collections.abc import Callable
from dataclasses import fields, is_dataclass
from typing import (
    Any,
    ClassVar,
    Protocol,
    TypeVar,
    get_type_hints,
    overload,
    runtime_checkable,
)

from aiovantage.command_client import CommandClient, Converter
from aiovantage.errors import NotImplementedError, NotSupportedError

T = TypeVar("T")


class _AsyncCallable(Protocol):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


@runtime_checkable
class _MethodCallable(Protocol):
    method_metadata: list[tuple[str, str | None, str | None, bool]]

    async def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


def method(
    *methods: str,
    out: str | None = None,
    property: str | None = None,
    fetch: bool = True,
) -> Callable[[T], T]:
    """Decorator to annotate a function as a Vantage method.

    This is used to automatically keep track of expected return types for
    Vantage method calls so we can parse responses correctly, both when
    directly invoking methods and when receiving status messages.

    Optionally, a property name can be associated with the method, which can
    be used to update the state of the object when receiving status messages,
    or to fetch the initial state of the object.

    Args:
        methods: The vantage method name(s) to associate with the function.
        out: Optional source of the return value, either "return" or "argN".
        property: Optional property name to associate with the function.
        fetch: Whether to fetch the property when fetching state.
    """

    def decorator(func: T) -> T:
        # Attach metadata to the function
        metadata: list[tuple[str, str | None, str | None, bool]] = getattr(
            func, "method_metadata", []
        )

        for method in methods:
            metadata.append((method, out, property, fetch))

        func.method_metadata = metadata  # type: ignore

        return func

    return decorator


class _InterfaceMeta(type):
    """Metaclass to collect method metadata from member functions and base classes."""

    def __new__(
        cls: type["_InterfaceMeta"],
        name: str,
        bases: tuple[type, ...],
        dct: dict[str, Any],
    ):
        """Create a new object interface class."""
        method_signatures: dict[str, type[Any]] = {}
        method_output: dict[str, str] = {}
        method_properties: dict[str, str] = {}
        property_getters: dict[str, _AsyncCallable] = {}

        # Include method metadata from base classes
        for base in bases:
            if issubclass(base, Interface):
                method_signatures.update(base._method_signatures)  # type: ignore
                method_output.update(base._method_output)  # type: ignore
                method_properties.update(base._method_properties)  # type: ignore
                property_getters.update(base._property_getters)  # type: ignore

        # Collect method metadata from member functions
        for attr in dct.values():
            if not isinstance(attr, _MethodCallable):
                continue

            for method, output, property, fetch in attr.method_metadata:
                # Get the return type of the method
                type_hints = get_type_hints(attr)
                if "return" not in type_hints:
                    raise NotImplementedError(f"Return type missing for {method}")

                # Get the fully-qualified method name
                fq_method = f"{dct['interface_name']}.{method}"

                # Save mapping between method name and return type
                # Used for parsing responses
                method_signatures[fq_method] = type_hints["return"]

                # Save mapping between method name and output argument
                # Used for parsing responses
                if output:
                    method_output[fq_method] = output

                # Save mapping between method name and property name
                # Used to update properties when receiving status messages
                if property:
                    method_properties[fq_method] = property

                # Save mapping between property name and getter function
                # Used to fetch state properties
                if property and fetch:
                    property_getters[property] = attr

        # Attach the method metadata to the class
        dct["_method_signatures"] = method_signatures
        dct["_method_output"] = method_output
        dct["_method_properties"] = method_properties
        dct["_property_getters"] = property_getters

        return super().__new__(cls, name, bases, dct)


class Interface(metaclass=_InterfaceMeta):
    """Base class for object interfaces."""

    interface_name: ClassVar[str]
    """The name of the interface."""

    command_client: CommandClient | None = None
    """The command client instance to use for making requests."""

    vid: int
    """The Vantage ID of the object to send requests to, typically set in a subclass."""

    # Method metadata
    _method_signatures: dict[str, type[Any]]
    _method_output: dict[str, str]
    _method_properties: dict[str, str]
    _property_getters: dict[str, _AsyncCallable]

    @overload
    async def invoke(self, method: str, *params: Any) -> Any: ...

    @overload
    async def invoke(self, method: str, *params: Any, as_type: type[T]) -> T: ...

    async def invoke(
        self, method: str, *params: Any, as_type: type[T] | None = None
    ) -> T | Any:
        """Invoke a method on an object, and return the parsed response.

        Args:
            method: The method to invoke.
            params: The parameters to send with the method.
            as_type: The expected return type of the method, will attempt to infer if not provided.

        Returns:
            A parsed response, or None if no response was expected.
        """
        # Make sure we have a command client to send requests with
        if not self.command_client:
            raise ValueError("The object has no command client to send requests with.")

        # Build the request
        request = f"INVOKE {self.vid} {method}"
        if params:
            request += " " + " ".join(Converter.serialize(p) for p in params)

        # Send the request
        response = await self.command_client.raw_request(request)

        # Break the response into tokens
        return_line = response[-1]
        _command, _vid, result, _method, *args = Converter.tokenize(return_line)

        # Parse the response
        return self._parse_object_response(method, result, *args, as_type=as_type)

    def update_properties(self, properties: dict[str, Any]) -> list[str]:
        """Update object properties.

        Args:
            properties: A dictionary of property names and their new values.

        Returns:
            A list of property names that were updated.
        """
        changed: list[str] = []
        for prop, value in properties.items():
            if hasattr(self, prop) and getattr(self, prop) != value:
                setattr(self, prop, value)
                changed.append(prop)

        return changed

    async def fetch_state(self) -> list[str]:
        """Fetch state properties provided by the interface(s) this object implements.

        Returns:
            A list of property names that were updated.
        """
        # cls = type(self)

        # Fetch each state property
        fetched_properties: dict[str, Any] = {}
        for prop, getter in self._property_getters.items():
            try:
                fetched_properties[prop] = await getter(self)
            except (NotImplementedError, NotSupportedError):
                continue

        return self.update_properties(fetched_properties)

    def handle_object_status(self, method: str, result: str, *args: str) -> list[str]:
        """Handle an object interface status message.

        Args:
            method: The method that was invoked.
            result: The result of the command.
            args: The arguments that were sent with the command.

        Returns:
            A list of property names that were updated.
        """
        # Look up the property associated with this method
        property = self._method_properties.get(method)
        if property is None:
            return []

        # Parse the response and update the property
        return self.update_properties(
            {property: self._parse_object_response(method, result, *args)}
        )

    def handle_category_status(self, category: str, *args: str) -> list[str]:
        """Handle category status messages.

        Object interfaces which can handle "legacy" status messages from the
        Host Command service should override this method.

        Args:
            category: The category of the status message, eg. "LOAD".
            args: The arguments that were sent with the command.

        Returns:
            A list of property names that were updated.
        """
        return []

    @classmethod
    def _parse_object_response(
        cls, method: str, result: str, *args: str, as_type: type[T] | None = None
    ) -> T | Any:
        # Parse an object interface response message, either from the response
        # to an INVOKE command, or from a status message.

        # -> R:INVOKE <id> <result> <method> <arg1> <arg2> ...
        # -> EL: <id> <method> <result> <arg1> <arg2> ...
        # -> S:STATUS <id> <method> <result> <arg1> <arg2> ...

        # Make sure we have a signature for the method
        if as_type is None and method not in cls._method_signatures:
            raise NotImplementedError(f"No signature found for method {method}")

        # Get the expected return type of the method, return early if None
        signature = as_type or cls._method_signatures[method]
        if signature in (None, type(None)):
            return None

        # Grab either the result or an argument based on "out" metadata field
        def get_output_value(out: str) -> Any:
            if out == "return":
                return result

            if out.startswith("arg") and out[3:].isdigit():
                index = int(out[3:])
                if 0 <= index < len(args):
                    return args[index]

            raise ValueError(f"Invalid 'out' metadata when parsing {method}")

        # Parse the response based on the expected return type
        if is_dataclass(signature):
            # If the method returns a dataclass, parse the result and/or arguments
            # into the expected fields of the dataclass, based on the field metadata.
            type_hints = get_type_hints(signature)
            props: dict[str, Any] = {}

            for field in fields(signature):
                out = field.metadata.get("out")
                if out is None:
                    raise ValueError(f"Field {field.name} missing 'out' metadata")

                field_signature = type_hints.get(field.name)
                if field_signature is None:
                    raise ValueError(f"Field {field.name} missing type hint")

                props[field.name] = Converter.deserialize(
                    field_signature, get_output_value(out)
                )

            return signature(**props)
        else:
            # Otherwise, parse the result into the expected type
            out = cls._method_output.get(method, "return")
            return Converter.deserialize(signature, get_output_value(out))



import asyncio
from collections.abc import Callable
from dataclasses import fields
from enum import Enum
from typing import TYPE_CHECKING, TypeVar, cast

from aiovantage._logger import logger
from aiovantage.command_client import Converter
from aiovantage.config_client import ConfigurationInterface
from aiovantage.events import (
    EnhancedLogReceived,
    EventDispatcher,
    ObjectAdded,
    ObjectDeleted,
    ObjectUpdated,
    Reconnected,
    StatusReceived,
)
from aiovantage.objects import SystemObject

from .query import QuerySet

if TYPE_CHECKING:
    from aiovantage import Vantage

T = TypeVar("T", bound=SystemObject)


class StatusType(Enum):
    """The type of status that the controller is using for state monitoring."""

    OBJECT = 1
    """Object status events from the Enhanced Log."""

    CATEGORY = 2
    """Category status events, eg: S:LOAD, S:BLIND, etc."""


class Controller(QuerySet[T], EventDispatcher):
    """Base controller for managing collections of Vantage objects."""

    vantage_types: tuple[str, ...]
    """The Vantage object types that this controller will fetch."""

    force_category_status: bool = False
    """Whether to force the controller to handle 'STATUS' categories."""

    def __init__(self, vantage: "Vantage") -> None:
        """Initialize a controller.

        Args:
            vantage: The Vantage instance.
        """
        self._vantage = vantage
        self._objects: dict[int, T] = {}
        self._initialized = False
        self._status_type: StatusType | None = None
        self._status_unsubs: list[Callable[[], None]] = []
        self._lock = asyncio.Lock()

        QuerySet[T].__init__(self, self._objects, self._lazy_initialize)
        EventDispatcher.__init__(self)

    def __getitem__(self, vid: int) -> T:
        """Return the object with the given Vantage ID."""
        return self._objects[vid]

    def __contains__(self, vid: int) -> bool:
        """Return True if the object with the given Vantage ID exists."""
        return vid in self._objects

    @property
    def status_type(self) -> StatusType | None:
        """Return the type of status event that the controller is monitoring."""
        return self._status_type

    async def initialize(
        self, *, fetch_state: bool = True, enable_state_monitoring: bool = True
    ) -> None:
        """Populate the controller, and optionally fetch object state.

        Args:
            fetch_state: Whether to fetch the state properties of objects.
            enable_state_monitoring: Whether to monitor for state changes on objects.
        """
        # Prevent concurrent controller initialization from multiple tasks, since we
        # are batch-modifying the _items dict.
        async with self._lock:
            prev_ids = set(self._objects.keys())
            cur_ids: set[int] = set()

            # Fetch all objects managed by this controller
            async for obj in ConfigurationInterface.get_objects(
                self._vantage.config_client, *self.vantage_types, as_type=SystemObject
            ):
                obj = cast(T, obj)

                if obj.vid in prev_ids:
                    # This is an existing object.
                    existing_obj = self._objects[obj.vid]

                    # Check if any attributes have changed and update them
                    attrs_changed: list[str] = []
                    for f in fields(type(obj)):
                        if hasattr(existing_obj, f.name):
                            new_value = getattr(obj, f.name)
                            if getattr(existing_obj, f.name) != new_value:
                                setattr(existing_obj, f.name, new_value)
                                attrs_changed.append(f.name)

                    # Notify subscribers if any attributes changed
                    if attrs_changed:
                        self.emit(ObjectUpdated(existing_obj, attrs_changed))
                else:
                    # This is a new object.

                    # Attach the command client to the object
                    obj.command_client = self._vantage.command_client

                    # Add it to the controller and notify subscribers
                    self._objects[obj.vid] = obj
                    self.emit(ObjectAdded(obj))

                # Keep track of which objects we've seen
                cur_ids.add(obj.vid)

            # Handle objects that were removed
            for vid in prev_ids - cur_ids:
                obj = self._objects.pop(vid)
                self.emit(ObjectDeleted(obj))

        logger.info(
            "%s populated (%d objects)", type(self).__name__, len(self._objects)
        )

        # Mark the controller as initialized
        if not self._initialized:
            self._initialized = True

        # Fetch state and subscribe to state changes if requested
        if self._objects:
            if fetch_state:
                await self.fetch_state()

            if enable_state_monitoring:
                await self.enable_state_monitoring()

    async def fetch_state(self) -> None:
        """Fetch the state properties of all objects managed by this controller."""
        for obj in self._objects.values():
            # Fetch state, and notify subscribers if any attributes changed
            attrs_changed = await obj.fetch_state()
            if attrs_changed:
                self.emit(ObjectUpdated(obj, attrs_changed))

        logger.info("%s fetched state", type(self).__name__)

    async def enable_state_monitoring(self) -> None:
        """Monitor for state changes on objects managed by this controller."""
        if self._status_type is not None:
            return

        # Start the event stream if it isn't already running
        event_conn = await self._vantage.event_stream.start()

        # When available, we'll use "object" status events (subscribed via
        # the Enhanced Log) because they support a richer set of status properties.
        # If these are not supported—either due to older firmware, or if the
        # controller explicitly requesting category statuses, we'll fall back to
        # "category" status events.
        if event_conn.supports_enhanced_log and not self.force_category_status:
            # Subscribe to "object status" events from the Enhanced Log.
            status_unsub = self._vantage.event_stream.subscribe_enhanced_log(
                self._handle_enhanced_log_event, "STATUS", "STATUSEX"
            )

            self._status_type = StatusType.OBJECT
        else:
            # Subscribe to "STATUS {category}" updates
            status_unsub = self._vantage.event_stream.subscribe_status(
                self._handle_status_event
            )

            self._status_type = StatusType.CATEGORY

        # Subscribe to reconnect events from the event stream
        reconnect_unsub = self._vantage.event_stream.subscribe(
            Reconnected, self._handle_reconnect_event
        )

        # Keep track of the subscriptions so we can unsubscribe later
        self._status_unsubs.extend([status_unsub, reconnect_unsub])

        logger.info("%s subscribed to state changes", type(self).__name__)

    async def disable_state_monitoring(self) -> None:
        """Stop monitoring for state changes on objects managed by this controller."""
        if self._status_type is None:
            return

        # Unsubscribe status and reconnect events
        while self._status_unsubs:
            unsub = self._status_unsubs.pop()
            unsub()

        self._status_type = None

        logger.info("%s unsubscribed from state changes", type(self).__name__)

    def _handle_status_event(self, event: StatusReceived) -> None:
        # Look up the object that this event is for
        obj = self._objects.get(event.vid)
        if obj is None:
            return

        # Handle the event
        if event.category == "STATUS":
            # Handle "object interface" status events of the form:
            # -> S:STATUS <vid> <method> <result> <arg1> <arg2> ...
            attrs_changed = obj.handle_object_status(*event.args)
        else:
            # Handle "category" status events, eg: S:LOAD, S:BLIND, etc
            # -> S:LOAD <vid> <arg1> <arg2> ...
            attrs_changed = obj.handle_category_status(event.category, *event.args)

        # Notify subscribers if any attributes changed
        if attrs_changed:
            self.emit(ObjectUpdated(obj, attrs_changed))

    def _handle_enhanced_log_event(self, event: EnhancedLogReceived) -> None:
        # Tokenize STATUS/STATUSEX logs from the enhanced log.
        # These are "object interface" status messages, of the form:
        # -> EL: <vid> <method> <result> <arg1> <arg2> ...
        vid_str, method, result, *args = Converter.tokenize(event.log)

        # Pass the event to the controller, if this object is managed by it
        obj = self._objects.get(int(vid_str))
        if obj is None:
            return

        # Handle the event, and notify subscribers if any attributes changed
        attrs_changed = obj.handle_object_status(method, result, *args)
        if attrs_changed:
            self.emit(ObjectUpdated(obj, attrs_changed))

    def _handle_reconnect_event(self, event: Reconnected) -> None:
        # Fetch latest state if we've been disconnected
        asyncio.create_task(self.fetch_state())

    async def _lazy_initialize(self) -> None:
        # Initialize the controller if it isn't already initialized
        if not self._initialized:
            await self.initialize()



"""GMem (variable) object."""

from dataclasses import dataclass, field

from aiovantage.object_interfaces import GMemInterface

from .system_object import SystemObject
from .types import Array


@dataclass(kw_only=True)
class GMem(SystemObject, GMemInterface):
    """GMem (variable) object."""

    @dataclass
    class Data(Array):
        fixed: bool = field(default=False, metadata={"type": "Attribute"})

    @dataclass
    class Tag:
        type: str
        object: bool = field(
            default=False, metadata={"name": "object", "type": "Attribute"}
        )

    category: int
    data: Data = field(metadata={"name": "data"})
    persistent: bool = True
    tag: Tag

    @property
    def is_bool(self) -> bool:
        """Return True if GMem is boolean type."""
        return self.tag.type == "bool"

    @property
    def is_str(self) -> bool:
        """Return True if GMem is string type."""
        return self.tag.type == "Text"

    @property
    def is_int(self) -> bool:
        """Return True if GMem is integer type."""
        return self.tag.type in (
            "Delay",
            "DeviceUnits",
            "Level",
            "Load",
            "Number",
            "Seconds",
            "Task",
            "DegC",
        )

    @property
    def is_object_id(self) -> bool:
        """Return True if GMem is a pointer to an object."""
        return self.tag.object

    @property
    def is_fixed(self) -> bool:
        """Return True if GMem is a fixed point number."""
        return self.data.fixed



from decimal import Decimal

from typing_extensions import override

from .base import Interface, method


class TemperatureInterface(Interface):
    """Temperature interface."""

    interface_name = "Temperature"

    # Properties
    value: Decimal | None = None

    # Methods
    @method("GetValue", "GetValueHW", property="value")
    async def get_value(self, *, hw: bool = False) -> Decimal:
        """Get the value of a temperature sensor.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the temperature sensor, in degrees Celsius.
        """
        # INVOKE <id> Temperature.GetValue
        # -> R:INVOKE <id> <temp> Temperature.GetValue
        return await self.invoke(
            "Temperature.GetValueHW" if hw else "Temperature.GetValue"
        )

    @method("SetValue", "SetValueSW")
    async def set_value(self, value: Decimal, *, sw: bool = False) -> None:
        """Set the value of a temperature sensor.

        Args:
            value: The value to set the sensor to.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Temperature.SetValue <value>
        # -> R:INVOKE <id> <rcode> Temperature.SetValue <value>
        await self.invoke(
            "Temperature.SetValueSW" if sw else "Temperature.SetValue", value
        )

    @override
    def handle_category_status(self, category: str, *args: str) -> list[str]:
        if category == "TEMP":
            # STATUS TEMP
            # -> S:TEMP <id> <temp>
            return self.update_properties({"value": Decimal(args[0])})

        return super().handle_category_status(category, *args)



import datetime as dt
from enum import IntEnum, IntFlag

from .base import Interface, method


class ConfigurationInterface(Interface):
    """Configuration object interface."""

    interface_name = "Configuration"

    class Store(IntEnum):
        """Configuration store."""

        Unspecified = 0
        Datastore = 1
        Userstore = 2

    class Compression(IntFlag):
        """Configuration compression type."""

        none = 1
        zlib = 2
        snappy = 4

    class SolarEvent(IntEnum):
        """Solar event types."""

        Sunrise = 0
        Sunset = 1

    # Methods
    @method("GetControllerVID")
    async def get_controller_vid(self, controller: int) -> int:
        """Get the VID of a controller, based on the controller number.

        Args:
            controller: The controller number to get the VID of.
        """
        # INVOKE <id> Configuration.GetControllerVID <controller>
        # -> R:INVOKE <id> <rcode> Configuration.GetControllerVID <vid>
        return await self.invoke("Configuration.GetControllerVID", controller)

    @method("DeleteObject")
    async def delete_object(self, store: Store, vid: int) -> None:
        """Delete an object from the controller.

        Args:
            store: The store to delete the object from.
            vid: The VID of the object to delete.
        """
        # INVOKE <id> Configuration.DeleteObject <vid>
        # -> R:INVOKE <id> <rcode> Configuration.DeleteObject <vid>
        await self.invoke("Configuration.DeleteObject", store, vid)

    @method("CreateObject")
    async def create_object(self, type: str) -> int:
        """Create an object on the controller.

        Args:
            type: The type of object to create, eg. "Load".

        Returns:
            The VID of the created object, 0 if the object could not be created.
        """
        # INVOKE <id> Configuration.CreateObject <type>
        # -> R:INVOKE <id> <rcode> Configuration.CreateObject <type>
        return await self.invoke("Configuration.CreateObject", type)

    @method("GetModificationTime")
    async def get_modification_time(self) -> dt.datetime:
        """Get the modification time of this object.

        Returns:
            The modification time of the object, as a datetime object.
        """
        # INVOKE <id> Configuration.GetModificationTime
        # -> R:INVOKE <id> <mtime> Configuration.GetModificationTime
        return await self.invoke("Configuration.GetModificationTime")

    @method("GetLastDeleteTime")
    async def get_last_delete_time(self, store: Store) -> dt.datetime:
        """Get the time of the last object deletion.

        Args:
            store: The store to get the last deletion time of.

        Returns:
            The time of the last object deletion, as a datetime object.
        """
        # INVOKE <id> Configuration.GetLastDeleteTime <store>
        # -> R:INVOKE <id> <time> Configuration.GetLastDeleteTime <store>
        return await self.invoke("Configuration.GetLastDeleteTime", store)

    @method("GetLastClearTime")
    async def get_last_clear_time(self) -> dt.datetime:
        """Get the time of the last clear.

        Returns:
            The time of the last store clear, as a datetime object.
        """
        # INVOKE <id> Configuration.GetLastClearTime
        # -> R:INVOKE <id> <time> Configuration.GetLastClearTime
        return await self.invoke("Configuration.GetLastClearTime")

    @method("OpenFilter")
    async def open_filter(self, store: Store, types: str = "", xpath: str = "") -> int:
        """Open a filter on a store.

        Args:
            store: The store to open the filter on.
            types: An optional comma-separated list of object types to filter on.
            xpath: An optional xpath expression to filter on, eg. "/Load", "/*[@VID='12']"

        Returns:
            An integer "handle" representing the opened filter.
        """
        # INVOKE <id> Configuration.OpenFilter <store> <types> <xpath>
        # -> R:INVOKE <id> <handle> Configuration.OpenFilter <store> <types> <xpath>
        return await self.invoke("Configuration.OpenFilter", store, types, xpath)

    @method("GetNextObjectVID")
    async def get_next_object_vid(self, handle: int) -> int:
        """Get the VID of the next object in a filter.

        Args:
            handle: The filter handle to get the next object of.

        Returns:
            The VID of the next object in the filter, or 0 if there are no more objects.
        """
        # INVOKE <id> Configuration.GetNextObjectVID <handle>
        # -> R:INVOKE <id> <vid> Configuration.GetNextObjectVID <handle>
        return await self.invoke("Configuration.GetNextObjectVID", handle)

    @method("CloseFilter")
    async def close_filter(self, handle: int) -> None:
        """Close a filter.

        Args:
            handle: The filter handle to close.
        """
        # INVOKE <id> Configuration.CloseFilter <handle>
        # -> R:INVOKE <id> <rcode> Configuration.CloseFilter <handle>
        await self.invoke("Configuration.CloseFilter", handle)

    @method("FindLocalObject")
    async def find_local_object(self, vid: int) -> bool:
        """Find a "local" object by VID, i.e. an object managed by this object.

        Args:
            vid: The VID of the object to find.

        Returns:
            True if the object is found, False otherwise.
        """
        # INVOKE <id> Configuration.FindLocalObject <vid>
        # -> R:INVOKE <id> <found (0/1)> Configuration.FindLocalObject <vid>
        return await self.invoke("Configuration.FindLocalObject", vid)

    @method("GetTimeZone", out="arg0")
    async def get_time_zone(self) -> str:
        """Get the time zone.

        Returns:
            The vantage time zone string, eg. "UTCPlusEight+8DAY,M3.2.0/02:00,M11.1.0/02:00"
        """
        # INVOKE <id> Configuration.GetTimeZone
        # -> R:INVOKE <id> <rcode> Configuration.GetTimeZone <tz> <size>
        return await self.invoke("Configuration.GetTimeZone")

    @method("GetTimeLocation", out="arg0")
    async def get_time_location(self) -> str:
        """Get the time location.

        Returns:
            A latitude and longitude string, eg. "51.178908N1.826212W"
        """
        # INVOKE <id> Configuration.GetTimeLocation
        # -> R:INVOKE <id> <rcode> Configuration.GetTimeLocation <loc> <size>
        return await self.invoke("Configuration.GetTimeLocation")

    @method("GetAstronomicalTime")
    async def get_astronomical_time(
        self, event: SolarEvent, year: int, month: int, day: int
    ) -> dt.datetime:
        """Get the astronomical time of a solar event.

        Args:
            event: The solar event to get the time of.
            year: The year of the event.
            month: The month of the event.
            day: The day of the event.

        Returns:
            The time of the solar event, as a datetime object.
        """
        # INVOKE <id> Configuration.GetAstronomicalTime <event>
        # -> R:INVOKE <id> <time> Configuration.GetAstronomicalTime <event> <year> <month> <day>
        return await self.invoke(
            "Configuration.GetAstronomicalTime", event, year, month, day
        )



from decimal import Decimal
from enum import IntEnum

from .base import Interface, method


class SounderInterface(Interface):
    """Sounder interface."""

    interface_name = "Sounder"

    class Status(IntEnum):
        """Sounder status."""

        On = 0
        Off = 1

    # Methods
    @method("GetFrequency", "GetFrequencyHW")
    async def get_frequency(self, *, hw: bool = False) -> Decimal:
        """Get the frequency of the keypad speaker.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The frequency of the keypad speaker in Hz.
        """
        # INVOKE <id> Sounder.GetFrequency
        # -> R:INVOKE <id> <frequency> Sounder.GetFrequency
        return await self.invoke(
            "Sounder.GetFrequencyHW" if hw else "Sounder.GetFrequency"
        )

    @method("SetFrequency")
    async def set_frequency(self, frequency: float | Decimal) -> None:
        """Set the frequency of the keypad speaker.

        Args:
            frequency: The frequency to set the sounder to, in Hz.
        """
        # INVOKE <id> Sounder.SetFrequency <frequency>
        # -> R:INVOKE <id> <rcode> Sounder.SetFrequency
        await self.invoke("Sounder.SetFrequency", frequency)

    @method("GetDuration", "GetDurationHW")
    async def get_duration(self, *, hw: bool = False) -> Decimal:
        """Get the length of time the keypad speaker will sound.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The duration of the keypad speaker sound in seconds.
        """
        # INVOKE <id> Sounder.GetDuration
        # -> R:INVOKE <id> <duration> Sounder.GetDuration
        return await self.invoke(
            "Sounder.GetDurationHW" if hw else "Sounder.GetDuration"
        )

    @method("SetDuration")
    async def set_duration(self, duration: Decimal) -> None:
        """Set the length of time the keypad speaker will sound.

        Args:
            duration: The duration to set the sounder to, in seconds.
        """
        # INVOKE <id> Sounder.SetDuration <duration>
        # -> R:INVOKE <id> <rcode> Sounder.SetDuration
        await self.invoke("Sounder.SetDuration", duration)

    async def get_status(self, *, hw: bool = False) -> Status:
        """Get the status of the keypad speaker.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The status of the keypad speaker.
        """
        # INVOKE <id> Sounder.GetStatus
        # -> R:INVOKE <id> <status (0/1)> Sounder.GetStatus
        return await self.invoke("Sounder.GetStatusHW" if hw else "Sounder.GetStatus")

    async def set_status(self, status: Status) -> None:
        """Set the status of the keypad speaker.

        Args:
            status: The status to set the keypad speaker to.
        """
        # INVOKE <id> Sounder.SetStatus <status (0/1/On/Off)>
        # -> R:INVOKE <id> <rcode> Sounder.SetStatus
        await self.invoke("Sounder.SetStatus", status)

    async def play_fx(
        self, effect: int, duration: float = 0, volume: float = 0
    ) -> None:
        """Play a sound effect on the keypad speaker.

        Args:
            effect: The effect to play.
            duration: The duration to play the FX for, in seconds, 0 for default.
            volume: The volume to play the FX at, as a percentage, 0 for default.
        """
        # INVOKE <id> Sounder.PlayFX <fx> <duration> <volume>
        # -> R:INVOKE <id> <rcode> Sounder.PlayFX
        await self.invoke("Sounder.PlayFX", effect, duration, volume)

    # Convenience functions, not part of the interface
    async def turn_on(self) -> None:
        """Turn on the keypad speaker."""
        await self.set_status(self.Status.On)

    async def turn_off(self) -> None:
        """Turn off the keypad speaker."""
        await self.set_status(self.Status.Off)



from decimal import Decimal
from enum import IntEnum

from .base import Interface, method


class ColorTemperatureInterface(Interface):
    """Interface for querying and controlling color temperature."""

    interface_name = "ColorTemperature"

    class Preset(IntEnum):
        """Color temperature preset."""

        None_ = 0
        Soft = 2700
        Warm = 3000
        Neutral = 3500
        Cool = 4100
        Natural = 5000
        Daylight = 6500

    # Properties
    color_temp: int | None = None
    max_value: int | None = None
    min_value: int | None = None

    # Methods
    @method("Set", "SetSW")
    async def set_color_temp(
        self, temp: int, transition: int = 0, *, sw: bool = False
    ) -> None:
        """Set the color temperature of a light.

        Args:
            temp: The color temperature to set the light to, in Kelvin.
            transition: The time in seconds to transition to the new color temperature.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> ColorTemperature.Set <temp> <seconds>
        # -> R:INVOKE <id> <rcode> ColorTemperature.Set <temp>
        await self.invoke(
            "ColorTemperature.SetSW" if sw else "ColorTemperature.Set", temp, transition
        )

    @method("Get", "GetHW", property="color_temp")
    async def get_color_temp(self, *, hw: bool = False) -> int:
        """Get the color temperature of a light.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The color temperature of the light, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.Get
        # -> R:INVOKE <id> <temp> ColorTemperature.Get
        return await self.invoke(
            "ColorTemperature.GetHW" if hw else "ColorTemperature.Get"
        )

    @method("StopTransition")
    async def stop_transition(self) -> None:
        """Stop any ongoing color temperature transitions."""
        # INVOKE <id> ColorTemperature.StopTransition
        # -> R:INVOKE <id> <rcode> ColorTemperature.StopTransition
        await self.invoke("ColorTemperature.StopTransition")

    @method("Warm")
    async def warm(self, amount: int, transition_time: float | Decimal) -> None:
        """Decrease the color temperature of a light.

        Args:
            amount: The amount to decrease the color temperature by.
            transition_time: The time in seconds to transition to the new color.
        """
        # INVOKE <id> ColorTemperature.Warm <amount> <transition_time>
        # -> R:INVOKE <id> <rcode> ColorTemperature.Warm <amount> <transition_time>
        await self.invoke("ColorTemperature.Warm", amount, transition_time)

    @method("Cool")
    async def cool(self, amount: int, transition_time: float | Decimal) -> None:
        """Increase the color temperature of a light.

        Args:
            amount: The amount to increase the color temperature by.
            transition_time: The time in seconds to transition to the new color.
        """
        # INVOKE <id> ColorTemperature.Cool <amount> <transition_time>
        # -> R:INVOKE <id> <rcode> ColorTemperature.Cool <amount> <transition_time>
        await self.invoke("ColorTemperature.Cool", amount, transition_time)

    @method("SetPreset")
    async def set_temperature_preset(
        self, value: Preset, transition_time: float | Decimal
    ) -> None:
        """Set the color temperature of a light to a preset value.

        Args:
            value: The preset value to set the light to.
            transition_time: The time in seconds to transition to the new color.
        """
        # INVOKE <id> ColorTemperature.SetPreset <value> <transition_time>
        # -> R:INVOKE <id> <rcode> ColorTemperature.SetPreset <value> <transition_time>
        await self.invoke("ColorTemperature.SetPreset", value, transition_time)

    @method("GetPreset")
    async def get_temperature_preset(self) -> Preset:
        """Get the color temperature preset of a light.

        Returns:
            The color temperature preset of the light.
        """
        # INVOKE <id> ColorTemperature.GetPreset
        # -> R:INVOKE <id> <preset> ColorTemperature.GetPreset
        return await self.invoke("ColorTemperature.GetPreset")

    @method("GetMaxValue", "GetMaxValueHW", property="max_value")
    async def get_max_value(self, *, hw: bool = False) -> int:
        """Get the maximum color temperature of a light.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The maximum color temperature of the light, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.GetMaxValue
        # -> R:INVOKE <id> <temp> ColorTemperature.GetMaxValue
        return await self.invoke(
            "ColorTemperature.GetMaxValueHW" if hw else "ColorTemperature.GetMaxValue",
        )

    @method("SetMaxValueSW")
    async def set_max_value(self, value: int) -> None:
        """Set the cached maximum color temperature of a light.

        Args:
            value: The maximum color temperature to set the light to, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.SetMaxValueSW <value>
        # -> R:INVOKE <id> <rcode> ColorTemperature.SetMaxValueSW <value>
        await self.invoke("ColorTemperature.SetMaxValueSW", value)

    @method("GetMinValue", "GetMinValueHW", property="min_value")
    async def get_min_value(self, *, hw: bool = False) -> int:
        """Get the minimum color temperature of a light.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The minimum color temperature of the light, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.GetMinValue
        # -> R:INVOKE <id> <temp> ColorTemperature.GetMinValue
        return await self.invoke(
            "ColorTemperature.GetMinValueHW" if hw else "ColorTemperature.GetMinValue",
        )

    @method("SetMinValueSW")
    async def set_min_value(self, value: int) -> None:
        """Set the cached minimum color temperature of a light.

        Args:
            value: The minimum color temperature to set the light to, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.SetMinValueSW <value>
        # -> R:INVOKE <id> <rcode> ColorTemperature.SetMinValueSW <value>
        await self.invoke("ColorTemperature.SetMinValueSW", value)

    @method("GetTransitionTemperature")
    async def get_transition_temperature(self) -> int:
        """Get the current color temperature of a light in transition.

        Returns:
            The transition temperature of the light, in Kelvin.
        """
        # INVOKE <id> ColorTemperature.GetTransitionTemperature
        # -> R:INVOKE <id> <temp> ColorTemperature.GetTransitionTemperature
        return await self.invoke("ColorTemperature.GetTransitionTemperature")



from decimal import Decimal

from .base import Interface, method


class SensorInterface(Interface):
    """Sensor interface."""

    interface_name = "Sensor"

    # Properties
    level: Decimal | None = None

    # Methods
    @method("GetLevel", "GetLevelHW", property="level")
    async def get_level(self, *, hw: bool = False) -> Decimal:
        """Get the level of a sensor.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The level of the sensor.
        """
        # INVOKE <id> Sensor.GetLevel
        # -> R:INVOKE <id> <level (0-100)> Sensor.GetLevel
        return await self.invoke("Sensor.GetLevelHW" if hw else "Sensor.GetLevel")

    @method("SetLevel")
    async def set_level(self, level: Decimal) -> None:
        """Set the level of a sensor.

        Args:
            level: The level to set the sensor to.
        """
        # INVOKE <id> Sensor.SetLevel <level (0-100)>
        # -> R:INVOKE <id> <rcode> Sensor.SetLevel <level (0-100)>
        await self.invoke("Sensor.SetLevel", level)

    @method("GetHighRange")
    async def get_high_range(self) -> Decimal:
        """Get the high range of a sensor.

        Returns:
            The high range of the sensor.
        """
        # INVOKE <id> Sensor.GetHighRange
        # -> R:INVOKE <id> <high range> Sensor.GetHighRange
        return await self.invoke("Sensor.GetHighRange")

    @method("GetLowRange")
    async def get_low_range(self) -> Decimal:
        """Get the low range of a sensor.

        Returns:
            The low range of the sensor.
        """
        # INVOKE <id> Sensor.GetLowRange
        # -> R:INVOKE <id> <low range> Sensor.GetLowRange
        return await self.invoke("Sensor.GetLowRange")

    @method("GetHoldOnTime")
    async def get_hold_on_time(self) -> Decimal:
        """Get the hold on time of a sensor.

        Returns:
            The hold on time of the sensor, in seconds.
        """
        # INVOKE <id> Sensor.GetHoldOnTime
        # -> R:INVOKE <id> <hold on time> Sensor.GetHoldOnTime
        return await self.invoke("Sensor.GetHoldOnTime")

    @method("IsTracking")
    async def is_tracking(self) -> bool:
        """Get whether the sensor is tracking.

        Returns:
            Whether the sensor is tracking.
        """
        # INVOKE <id> Sensor.IsTracking
        # -> R:INVOKE <id> <tracking (0/1)> Sensor.IsTracking
        return await self.invoke("Sensor.IsTracking")

    @method("GetTrackingDelta")
    async def get_tracking_delta(self) -> Decimal:
        """Get the tracking delta of a sensor.

        Returns:
            The tracking delta of the sensor.
        """
        # INVOKE <id> Sensor.GetTrackingDelta
        # -> R:INVOKE <id> <tracking delta> Sensor.GetTrackingDelta
        return await self.invoke("Sensor.GetTrackingDelta")

    @method("GetTrackingMin")
    async def get_tracking_min(self) -> Decimal:
        """Get the tracking min time of a sensor.

        Returns:
            The tracking min time of the sensor, in seconds.
        """
        # INVOKE <id> Sensor.GetTrackingMin
        # -> R:INVOKE <id> <tracking min> Sensor.GetTrackingMin
        return await self.invoke("Sensor.GetTrackingMin")

    @method("GetTrackingMax")
    async def get_tracking_max(self) -> Decimal:
        """Get the tracking max time of a sensor.

        Returns:
            The tracking max time of the sensor, in seconds.
        """
        # INVOKE <id> Sensor.GetTrackingMax
        # -> R:INVOKE <id> <tracking max> Sensor.GetTrackingMax
        return await self.invoke("Sensor.GetTrackingMax")



"""DIN Contact Input Station."""

from dataclasses import dataclass

from .din_station import DINStation


@dataclass(kw_only=True)
class DINContactInput(DINStation):
    """DIN Contact Input Station."""



"""BlindGroup object."""

from dataclasses import dataclass, field

from aiovantage.object_interfaces import BlindInterface

from .location_object import LocationObject


@dataclass(kw_only=True)
class BlindGroup(LocationObject, BlindInterface):
    """BlindGroup object."""

    category: int
    blind_table: list[int] = field(
        default_factory=list,
        metadata={
            "name": "Blind",
            "wrapper": "BlindTable",
        },
    )



from decimal import Decimal

from typing_extensions import override

from .base import Interface, method


class LightSensorInterface(Interface):
    """Light sensor object interface."""

    interface_name = "LightSensor"

    # Properties
    level: Decimal | None = None

    # Methods
    @method("GetLevel", "GetLevelHW", property="level")
    async def get_level(self, *, hw: bool = False) -> Decimal:
        """Get the level of a light sensor.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The level of the light sensor, in foot-candles.
        """
        # INVOKE <id> LightSensor.GetLevel
        # -> R:INVOKE <id> <level> LightSensor.GetLevel
        return await self.invoke(
            "LightSensor.GetLevelHW" if hw else "LightSensor.GetLevel"
        )

    @method("SetLevel", "SetLevelSW")
    async def set_level(self, level: Decimal) -> None:
        """Set the level of a light sensor.

        Args:
            level: The level to set, in foot-candles.
        """
        # INVOKE <id> LightSensor.SetLevel <level>
        # -> R:INVOKE <id> <rcode> LightSensor.SetLevel <level>
        await self.invoke("LightSensor.SetLevel", level)

    @override
    def handle_category_status(self, category: str, *args: str) -> list[str]:
        if category == "LIGHT":
            # STATUS LIGHT
            # -> S:LIGHT <id> <level>
            return self.update_properties({"level": Decimal(args[0])})

        return super().handle_category_status(category, *args)



"""Master (controller) object."""

from dataclasses import dataclass, field

from aiovantage.object_interfaces import ConfigurationInterface, IntrospectionInterface

from .system_object import SystemObject


@dataclass(kw_only=True)
class Master(SystemObject, IntrospectionInterface, ConfigurationInterface):
    """Master (InFusion Controller) object."""

    @dataclass
    class DINEnclosure:
        enclosure: int
        position: int = field(metadata={"type": "Attribute"})
        row: int = field(metadata={"type": "Attribute"})

    number: int
    boot: int = 0
    volts: float = 24.0
    amps: float = 2.5

    # Not available in 2.x firmware
    din_enclosure: DINEnclosure | None = field(
        default=None,
        metadata={
            "name": "DINEnclosure",
        },
    )
    module_count: int | None = None
    power_supply: int | None = None
    serial_number: int | None = None



from aiovantage.objects import BlindGroup, SomfyRS485GroupChild, SomfyURTSI2GroupChild

from .base import Controller

BlindGroupTypes = BlindGroup | SomfyRS485GroupChild | SomfyURTSI2GroupChild
"""Types managed by the blind groups controller."""


class BlindGroupsController(Controller[BlindGroupTypes]):
    """Blind groups controller."""

    vantage_types = (
        "BlindGroup",
        "Somfy.RS-485_Group_CHILD",
        "Somfy.URTSI_2_Group_CHILD",
    )



"""Somfy RS-485 SDN 2.0 driver objects."""

from dataclasses import dataclass

from aiovantage.object_interfaces import BlindInterface

from .child_device import ChildDevice
from .port_device import PortDevice


@dataclass(kw_only=True)
class SomfyRS485SDN20Port(PortDevice):
    """Somfy RS-485 SDN 2.0 port device."""

    class Meta:
        name = "Somfy.RS-485_SDN_2_x2E_0_PORT"


@dataclass(kw_only=True)
class SomfyRS485ShadeChild(ChildDevice, BlindInterface):
    """Somfy RS-485 SDN 2.0 blind."""

    class Meta:
        name = "Somfy.RS-485_Shade_CHILD"


@dataclass(kw_only=True)
class SomfyRS485GroupChild(ChildDevice, BlindInterface):
    """Somfy RS-485 SDN 2.0 blind group."""

    class Meta:
        name = "Somfy.RS-485_Group_CHILD"



"""Area object."""

from dataclasses import dataclass

from .location_object import LocationObject


@dataclass(kw_only=True)
class Area(LocationObject):
    """Area object."""

    area_type: str
    enclosure_area: int | None = None



"""Temperature object."""

from dataclasses import dataclass

from aiovantage.object_interfaces import TemperatureInterface

from .sensor import Sensor
from .types import Parent


@dataclass(kw_only=True)
class Temperature(Sensor, TemperatureInterface):
    """Temperature object."""

    parent: Parent
    out_of_range: int = 0
    in_range: int = 0
    range_high: float
    range_low: float
    hold_on_time: float
    temp: int = 0



"""AnemoSensor (wind sensor) object."""

from dataclasses import dataclass

from aiovantage.object_interfaces import AnemoSensorInterface, SensorInterface

from .sensor import Sensor
from .types import Parent


@dataclass(kw_only=True)
class AnemoSensor(Sensor, SensorInterface, AnemoSensorInterface):
    """AnemoSensor (wind sensor) object."""

    parent: Parent
    out_of_range: int = 0
    in_range: int = 0
    range_high: float = 10.0
    range_low: float = 0.0
    hold_on_time: float



"""Fetch all omni sensors from the Vantage controller, and print any state changes."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage import Vantage
from aiovantage.events import ObjectUpdated
from aiovantage.objects import OmniSensor

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def on_object_updated(event: ObjectUpdated[OmniSensor]) -> None:
    """Print out any state changes."""
    print(f"[OmniSensor updated] '{event.obj.name}' ({event.obj.id})")
    for attr in event.attrs_changed:
        print(f"    {attr} = {getattr(event.obj, attr)}")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Fetch all known sensors from the controller
        await vantage.omni_sensors.initialize()

        # Subscribe to updates for all sensors
        vantage.omni_sensors.subscribe(ObjectUpdated, on_object_updated)

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Base class for child device (driver-provided) objects."""

from dataclasses import dataclass

from .custom_device import CustomDevice
from .types import Parent


@dataclass(kw_only=True)
class ChildDevice(CustomDevice):
    """Base class for child device (driver-provided) objects."""

    parent: Parent



from decimal import Decimal

from typing_extensions import override

from .base import Interface, method


class AnemoSensorInterface(Interface):
    """Anemo sensor object interface."""

    interface_name = "AnemoSensor"

    # Properties
    speed: Decimal | None = None

    # Methods
    @method("GetSpeed", "GetSpeedHW", property="speed")
    async def get_speed(self, *, hw: bool = False) -> Decimal:
        """Get the speed of an anemo sensor.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The speed of the anemo sensor, in mph.
        """
        # INVOKE <id> AnemoSensor.GetSpeed
        # -> R:INVOKE <id> <speed> AnemoSensor.GetSpeed
        return await self.invoke(
            "AnemoSensor.GetSpeedHW" if hw else "AnemoSensor.GetSpeed"
        )

    @method("SetSpeed", "SetSpeedSW")
    async def set_speed(self, speed: Decimal, *, sw: bool = False) -> None:
        """Set the speed of an anemo sensor.

        Args:
            speed: The speed to set, in mph.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> AnemoSensor.SetSpeed <speed>
        # -> R:INVOKE <id> <rcode> AnemoSensor.SetSpeed <speed>
        await self.invoke(
            "AnemoSensor.SetSpeedSW" if sw else "AnemoSensor.SetSpeed", speed
        )

    @override
    def handle_category_status(self, category: str, *args: str) -> list[str]:
        if category == "WIND":
            # STATUS WIND
            # -> S:WIND <id> <wind_speed>
            return self.update_properties({"speed": Decimal(args[0])})

        return super().handle_category_status(category, *args)



"""Keypad Station."""

from dataclasses import dataclass

from aiovantage.object_interfaces import SounderInterface

from .station_object import StationObject
from .types import Parent


@dataclass(kw_only=True)
class Keypad(StationObject, SounderInterface):
    """Keypad Station."""

    parent: Parent



from dataclasses import dataclass, field
from decimal import Decimal

from typing_extensions import override

from .base import Interface, method
from .fields import ShadeOrientation, ShadeType


class BlindInterface(Interface, ShadeOrientation, ShadeType):
    """Blind object interface."""

    interface_name = "Blind"

    @dataclass
    class BlindState:
        """The state of a blind."""

        is_moving: bool = field(metadata={"out": "return"})
        start_pos: Decimal = field(metadata={"out": "arg0"})
        end_pos: Decimal = field(metadata={"out": "arg1"})
        transition_time: Decimal = field(metadata={"out": "arg2"})
        start_time: int = field(metadata={"out": "arg3"})

    @dataclass
    class TravelTimes:
        """The travel times of a blind."""

        open_time: Decimal = field(metadata={"out": "arg0"})
        close_time: Decimal = field(metadata={"out": "arg1"})

    # Properties
    position: Decimal | None = None
    tilt_angle: int | None = None
    tilt_available: bool | None = None
    blind_state: BlindState | None = None

    # Methods
    @method("Open")
    async def open(self) -> None:
        """Open a blind."""
        # INVOKE <id> Blind.Open
        # -> R:INVOKE <id> <rcode> Blind.Open
        await self.invoke("Blind.Open")

    @method("Close")
    async def close(self) -> None:
        """Close a blind."""
        # INVOKE <id> Blind.Close
        # -> R:INVOKE <id> <rcode> Blind.Close
        await self.invoke("Blind.Close")

    @method("Stop")
    async def stop(self) -> None:
        """Stop a blind."""
        # INVOKE <id> Blind.Stop
        # -> R:INVOKE <id> <rcode> Blind.Stop
        await self.invoke("Blind.Stop")

    @method("SetPosition", "SetPositionSW")
    async def set_position(self, position: float, *, sw: bool = False) -> None:
        """Set the position of a blind.

        Args:
            position: The position to set the blind to, as a percentage.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetPosition <position>
        # -> R:INVOKE <id> <rcode> Blind.SetPosition <position>
        await self.invoke(
            "Blind.SetPositionSW" if sw else "Blind.SetPosition", position
        )

    @method("GetPosition", "GetPositionHW", property="position")
    async def get_position(self, *, hw: bool = False) -> Decimal:
        """Get the position of a blind.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The position of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetPosition
        # -> R:INVOKE <id> <position (0-100.000)> Blind.GetPosition
        return await self.invoke("Blind.GetPositionHW" if hw else "Blind.GetPosition")

    @method("SetTiltAngle", "SetTiltAngleSW")
    async def set_tilt_angle(self, angle: int, *, sw: bool = False) -> None:
        """Set the tilt angle of a blind.

        Args:
            angle: The angle to set the blind to, from -100 to 100.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetTiltAngle <angle>
        # -> R:INVOKE <id> <rcode> Blind.SetTiltAngle <angle>
        await self.invoke("Blind.SetTiltAngleSW" if sw else "Blind.SetTiltAngle", angle)

    @method("GetTiltAngle", "GetTiltAngleHW", property="tilt_angle")
    async def get_tilt_angle(self, *, hw: bool = False) -> int:
        """Get the tilt angle of a blind.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The tilt angle of the blind, from -100 to 100.
        """
        # INVOKE <id> Blind.GetTiltAngle
        # -> R:INVOKE <id> <angle (-100-100)> Blind.GetTiltAngle
        return await self.invoke("Blind.GetTiltAngleHW" if hw else "Blind.GetTiltAngle")

    @method("TiltClockwise")
    async def tilt_clockwise(self, angle: int) -> None:
        """Tilt the blinds clockwise by the specified angle.

        Args:
            angle: The angle offset the blinds should be tilted.
        """
        # INVOKE <id> Blind.TiltClockwise <angle>
        # -> R:INVOKE <id> <rcode> Blind.TiltClockwise <angle>
        await self.invoke("Blind.TiltClockwise", angle)

    @method("TiltCounterClockwise")
    async def tilt_counter_clockwise(self, angle: int) -> None:
        """Tilt the shades counter-clockwise by the specified angle.

        Args:
            angle: The angle offset the blinds should be tilted.
        """
        # INVOKE <id> Blind.TiltCounterClockwise <angle>
        # -> R:INVOKE <id> <rcode> Blind.TiltCounterClockwise <angle>
        await self.invoke("Blind.TiltCounterClockwise", angle)

    @method("IsTiltAvailable", property="tilt_available")
    async def is_tilt_available(self) -> bool:
        """Check if the blind can tilt in its current state.

        Returns:
            Whether the blind supports tilting.
        """
        # INVOKE <id> Blind.IsTiltAvailable
        # -> R:INVOKE <id> <available (0/1)> Blind.IsTiltAvailable
        return await self.invoke("Blind.IsTiltAvailable")

    @method("SetTiltAvailableSW")
    async def set_tilt_available(self, available: bool) -> None:
        """Set the cached tilt availability of a blind.

        Args:
            available: Whether the blind supports tilting.
        """
        # INVOKE <id> Blind.SetTiltAvailableSW <available>
        # -> R:INVOKE <id> <rcode> Blind.SetTiltAvailableSW <available>
        await self.invoke("Blind.SetTiltAvailableSW", available)

    @method("GetBlindState", property="blind_state")
    async def get_blind_state(self) -> BlindState:
        """Get the state of a blind.

        Returns:
            The state of the blind.
        """
        # INVOKE <id> Blind.GetBlindState
        # -> R:INVOKE <id> <moving> Blind.GetBlindState <start> <end> <transitionTime> <startTime>
        return await self.invoke("Blind.GetBlindState")

    @method("SetUpperLimit", "SetUpperLimitSW")
    async def set_upper_limit(self, limit: Decimal, *, sw: bool = False) -> None:
        """Set the upper limit of a blind.

        Args:
            limit: The upper limit to set the blind to, as a percentage.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetUpperLimit <limit>
        # -> R:INVOKE <id> <rcode> Blind.SetUpperLimit <limit>
        await self.invoke(
            "Blind.SetUpperLimitSW" if sw else "Blind.SetUpperLimit", limit
        )

    @method("GetUpperLimit", "GetUpperLimitHW")
    async def get_upper_limit(self, *, hw: bool = False) -> Decimal:
        """Get the upper limit of a blind.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The upper limit of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetUpperLimit
        # -> R:INVOKE <id> <limit (0-100.000)> Blind.GetUpperLimit
        return await self.invoke(
            "Blind.GetUpperLimitHW" if hw else "Blind.GetUpperLimit"
        )

    @method("SetLowerLimit", "SetLowerLimitSW")
    async def set_lower_limit(self, limit: Decimal, *, sw: bool = False) -> None:
        """Set the lower limit of a blind.

        Args:
            limit: The lower limit to set the blind to, as a percentage.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> Blind.SetLowerLimit <limit>
        # -> R:INVOKE <id> <rcode> Blind.SetLowerLimit <limit>
        await self.invoke(
            "Blind.SetLowerLimitSW" if sw else "Blind.SetLowerLimit", limit
        )

    @method("GetLowerLimit", "GetLowerLimitHW")
    async def get_lower_limit(self, *, hw: bool = False) -> Decimal:
        """Get the lower limit of a blind.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The lower limit of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetLowerLimit
        # -> R:INVOKE <id> <limit (0-100.000)> Blind.GetLowerLimit
        return await self.invoke(
            "Blind.GetLowerLimitHW" if hw else "Blind.GetLowerLimit"
        )

    @method("GetTravelTimes")
    async def get_travel_times(self) -> TravelTimes:
        """Get the travel times of a blind.

        Returns:
            The travel times of the blind.
        """
        # INVOKE <id> Blind.GetTravelTimes
        # -> R:INVOKE <id> <openTime> <closeTime> Blind.GetTravelTimes
        return await self.invoke("Blind.GetTravelTimes")

    @override
    def handle_category_status(self, category: str, *args: str) -> list[str]:
        if category == "BLIND":
            # STATUS BLIND
            # -> S:BLIND <id> <position (0.000 - 100.000)>
            return self.update_properties({"position": Decimal(args[0])})

        return super().handle_category_status(category, *args)



"""Vantage HVAC-IU objects."""

from dataclasses import dataclass, field

from aiovantage.object_interfaces import FanInterface, ThermostatInterface

from .child_device import ChildDevice
from .port_device import PortDevice


@dataclass(kw_only=True)
class VantageHVACIUPort(PortDevice):
    """Vantage HVAC-IU port device."""

    class Meta:
        name = "Vantage.HVAC-IU_PORT"

    temperature_format: str = field(
        default="Celcius",  # NOTE: Intentional typo to match the underlying object
        metadata={"name": "aTemperatureFormat"},
    )
    outdoor_sensor: int
    pauze_time: int = 1  # NOTE: Intentional typo to match the underlying object
    serial_number: str = "0"


@dataclass(kw_only=True)
class VantageHVACIULineChild(ChildDevice):
    """Vantage HVAC-IU line child device."""

    class Meta:
        name = "Vantage.HVAC-IU-Line_CHILD"

    @dataclass(kw_only=True)
    class OperationModes:
        auto: bool = True
        cool: bool = True
        heat: bool = True

    @dataclass(kw_only=True)
    class FanSpeeds:
        auto: bool = True
        high: bool = True
        low: bool = True
        max: bool = True
        med: bool = True

    device_type: str = "Daikin"
    line_number: int = 1
    operation_modes: OperationModes
    fan_speeds: FanSpeeds = field(metadata={"name": "xFanSpeeds"})


@dataclass(kw_only=True)
class VantageHVACIUZoneChild(ChildDevice, ThermostatInterface, FanInterface):
    """Vantage HVAC-IU zone child device."""

    class Meta:
        name = "Vantage.HVAC-IU-Zone_CHILD"

    @dataclass(kw_only=True)
    class IndoorSensor:
        indoor_sensor: int
        indoor_temp_offset: str = "0"

    main_zone: str = field(
        metadata={"name": "ZoneNumber", "wrapper": "aMainZonePlaceHolder"}
    )

    grouped_zones: list[str] = field(
        default_factory=list,
        metadata={"name": "ZoneNumberChild", "wrapper": "bGroupZonePlaceHolder"},
    )

    indoor_sensors: list[IndoorSensor] = field(
        default_factory=list, metadata={"name": "cIndoorSensors"}
    )



"""Vantage object definitions.

This module provides a (non-exhaustive) collection of classes that represent the
various object types that are managed by a Vantage system.

Vantage objects define *configuration* properties, which are properties that are
set during system programming, through Design Center.

Objects are hierarchical, with all objects ultimately deriving from
[`SystemObject`][aiovantage.objects.SystemObject].

Each object implements one or more [object interfaces][aiovantage.object_interfaces],
which provide *state* properties and methods for interacting with the object.
"""

from ._objects.anemo_sensor import AnemoSensor
from ._objects.area import Area
from ._objects.back_box import BackBox
from ._objects.blind import Blind
from ._objects.blind_group import BlindGroup
from ._objects.button import Button
from ._objects.child_device import ChildDevice
from ._objects.contact_input import ContactInput
from ._objects.custom_device import CustomDevice
from ._objects.dc_power_profile import DCPowerProfile
from ._objects.dimmer import Dimmer
from ._objects.din_contact_input import DINContactInput
from ._objects.din_low_voltage_relay_station import DINLowVoltageRelayStation
from ._objects.din_station import DINStation
from ._objects.dry_contact import DryContact
from ._objects.dual_relay_station import DualRelayStation
from ._objects.eq_ctrl import EqCtrl
from ._objects.eq_ux import EqUX
from ._objects.gmem import GMem
from ._objects.high_voltage_relay_station import HighVoltageRelayStation
from ._objects.irx2 import IRX2
from ._objects.keypad import Keypad
from ._objects.light_sensor import LightSensor
from ._objects.load import Load
from ._objects.load_group import LoadGroup
from ._objects.location_object import LocationObject
from ._objects.low_voltage_relay_station import LowVoltageRelayStation
from ._objects.master import Master
from ._objects.module import Module
from ._objects.module_gen2 import ModuleGen2
from ._objects.omni_sensor import OmniSensor
from ._objects.parent_device import ParentDevice
from ._objects.port_device import PortDevice
from ._objects.power_profile import PowerProfile
from ._objects.pwm_power_profile import PWMPowerProfile
from ._objects.qis_blind import QISBlind
from ._objects.qube_blind import QubeBlind
from ._objects.relay_blind import RelayBlind
from ._objects.rs232_station import RS232Station
from ._objects.rs485_station import RS485Station
from ._objects.scene_point_relay import ScenePointRelay
from ._objects.sensor import Sensor
from ._objects.somfy_rs_485_sdn_20 import (
    SomfyRS485GroupChild,
    SomfyRS485SDN20Port,
    SomfyRS485ShadeChild,
)
from ._objects.somfy_urtsi_2 import (
    SomfyURTSI2GroupChild,
    SomfyURTSI2Port,
    SomfyURTSI2ShadeChild,
)
from ._objects.station_bus import StationBus
from ._objects.station_object import StationObject
from ._objects.system_object import SystemObject
from ._objects.task import Task
from ._objects.temperature import Temperature
from ._objects.thermostat import Thermostat
from ._objects.types import Parent
from ._objects.vantage_ddg_color_load import VantageDDGColorLoad
from ._objects.vantage_dg_color_load import VantageDGColorLoad
from ._objects.vantage_dmx_dali_gateway import VantageDmxDaliGateway
from ._objects.vantage_dmx_gateway import VantageDmxGateway
from ._objects.vantage_generic_hvac_rs485 import (
    VantageGenericHVACRS485CompoundChild,
    VantageGenericHVACRS485Port,
    VantageGenericHVACRS485TechContactsChild,
    VantageGenericHVACRS485ZoneChild,
    VantageGenericHVACRS485ZoneWithoutFanSpeedChild,
)
from ._objects.vantage_hvac_iu import (
    VantageHVACIULineChild,
    VantageHVACIUPort,
    VantageHVACIUZoneChild,
)
from ._objects.vantage_virtual_thermostat import VantageVirtualThermostatPort

__all__ = [
    "AnemoSensor",
    "Area",
    "BackBox",
    "Blind",
    "BlindGroup",
    "Button",
    "ChildDevice",
    "ContactInput",
    "CustomDevice",
    "DCPowerProfile",
    "Dimmer",
    "DINContactInput",
    "DINLowVoltageRelayStation",
    "DINStation",
    "DryContact",
    "DualRelayStation",
    "EqCtrl",
    "EqUX",
    "GMem",
    "HighVoltageRelayStation",
    "IRX2",
    "Keypad",
    "LightSensor",
    "Load",
    "LoadGroup",
    "LocationObject",
    "LowVoltageRelayStation",
    "Master",
    "Module",
    "ModuleGen2",
    "OmniSensor",
    "Parent",
    "ParentDevice",
    "PortDevice",
    "PowerProfile",
    "PWMPowerProfile",
    "QISBlind",
    "QubeBlind",
    "RelayBlind",
    "RS232Station",
    "RS485Station",
    "ScenePointRelay",
    "Sensor",
    "SomfyRS485GroupChild",
    "SomfyRS485SDN20Port",
    "SomfyRS485ShadeChild",
    "SomfyURTSI2GroupChild",
    "SomfyURTSI2Port",
    "SomfyURTSI2ShadeChild",
    "StationBus",
    "StationObject",
    "SystemObject",
    "Task",
    "Temperature",
    "Thermostat",
    "VantageDDGColorLoad",
    "VantageDGColorLoad",
    "VantageDmxDaliGateway",
    "VantageDmxGateway",
    "VantageGenericHVACRS485CompoundChild",
    "VantageGenericHVACRS485Port",
    "VantageGenericHVACRS485TechContactsChild",
    "VantageGenericHVACRS485ZoneChild",
    "VantageGenericHVACRS485ZoneWithoutFanSpeedChild",
    "VantageHVACIULineChild",
    "VantageHVACIUPort",
    "VantageHVACIUZoneChild",
    "VantageVirtualThermostatPort",
]



"""Example of using the CommandClient to subscribe to system log events."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage.command_client import EventStream
from aiovantage.events import Connected, EnhancedLogReceived

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def on_connected(event: Connected) -> None:
    """Print out the connection event."""
    print("Connected and monitoring for log events...")


def command_client_callback(event: EnhancedLogReceived) -> None:
    """Print out the log message for each event."""
    print(event.log)


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Create an EventStream client
    async with EventStream(args.host, args.username, args.password) as events:
        # Subscribe to connection events
        events.subscribe(Connected, on_connected)

        # Subscribe to system log events
        events.subscribe_enhanced_log(command_client_callback, "SYSTEM")

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Print a list of all loads that are currently on every 5 seconds."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage import Vantage

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    async with Vantage(args.host, args.username, args.password) as vantage:
        # Preload the loads from the controller
        await vantage.loads.initialize()

        # Print a list of all loads that are currently on every 5 seconds
        while True:
            on_loads = list(vantage.loads.on)
            print(f"{len(on_loads)} loads are ON")
            for load in on_loads:
                print(f"- {load.name}")
            print()

            await asyncio.sleep(5)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Load object."""

from dataclasses import dataclass
from decimal import Decimal

from aiovantage.object_interfaces import LoadInterface

from .location_object import LocationObject
from .types import Parent


@dataclass(kw_only=True)
class Load(LocationObject, LoadInterface):
    """Load object."""

    parent: Parent
    contractor_number: str
    load_type: str = "Incandescent"
    power: int = 100
    power_profile: int
    override_level: Decimal = Decimal("100")

    @property
    def is_relay(self) -> bool:
        """Return True if the load type is a relay."""
        return self.load_type in (
            "High Voltage Relay",
            "Low Voltage Relay",
            "[MDR8RW101]",
        )

    @property
    def is_motor(self) -> bool:
        """Return True if the load type is a motor."""
        return self.load_type == "Motor"

    @property
    def is_light(self) -> bool:
        """Return True if the load type is inferred to be a light."""
        return not (self.is_relay or self.is_motor)

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)



from aiovantage.objects import VantageDDGColorLoad, VantageDGColorLoad

from .base import Controller
from .query import QuerySet

RGBLoadTypes = VantageDDGColorLoad | VantageDGColorLoad
"""Types managed by the RGB loads controller."""


class RGBLoadsController(Controller[RGBLoadTypes]):
    """RGB loads controller."""

    vantage_types = ("Vantage.DGColorLoad", "Vantage.DDGColorLoad")

    @property
    def on(self) -> QuerySet[RGBLoadTypes]:
        """Return a queryset of all RGB loads that are turned on."""
        return self.filter(lambda load: load.is_on)

    @property
    def off(self) -> QuerySet[RGBLoadTypes]:
        """Return a queryset of all RGB loads that are turned off."""
        return self.filter(lambda load: not load.is_on)



"""PWM power profile object."""

from dataclasses import dataclass

from .dc_power_profile import DCPowerProfile


@dataclass(kw_only=True)
class PWMPowerProfile(DCPowerProfile):
    """PWM power profile object."""

    inverted: bool = False



"""ScenePoint Dual Relay Station."""

from dataclasses import dataclass

from .keypad import Keypad


@dataclass(kw_only=True)
class DualRelayStation(Keypad):
    """ScenePoint Dual Relay Station."""

    shade_controller: bool
    reverse_shade: bool



from decimal import Decimal

from typing_extensions import override

from .base import Interface, method


class PowerSensorInterface(Interface):
    """Power sensor interface."""

    interface_name = "PowerSensor"

    # Properties
    power: Decimal | None = None

    # Methods
    @method("GetPower", "GetPowerHW", property="power")
    async def get_power(self, *, hw: bool = False) -> Decimal:
        """Get the value of a power sensor.

        Args:
            hw: Fetch the value from hardware instead of cache.

        Returns:
            The value of the power sensor, in Watts.
        """
        # INVOKE <id> PowerSensor.GetPower
        # -> R:INVOKE <id> <level> PowerSensor.GetPower
        return await self.invoke(
            "PowerSensor.GetPowerHW" if hw else "PowerSensor.GetPower"
        )

    @method("SetPower", "SetPowerSW")
    async def set_power(self, value: Decimal, *, sw: bool = False) -> None:
        """Set the value of a power sensor.

        Args:
            value: The value to set, in Watts.
            sw: Set the cached value instead of the hardware value.
        """
        # INVOKE <id> PowerSensor.SetPower <level>
        # -> R:INVOKE <id> <rcode> PowerSensor.SetPower <level>
        await self.invoke(
            "PowerSensor.SetPowerSW" if sw else "PowerSensor.SetPower", value
        )

    @override
    def handle_category_status(self, category: str, *args: str) -> list[str]:
        if category == "POWER":
            # STATUS POWER
            # -> S:POWER <id> <power>
            return self.update_properties({"power": Decimal(args[0])})

        return super().handle_category_status(category, *args)
